@@ -11,20 +11,29 @@ as taken from http://docs.python.org/dev/library/ssl.html#certificates
 
 '''
 
-import signal, socket, optparse, time, os, sys, subprocess, logging, errno
-try:    from socketserver import ForkingMixIn
-except: from SocketServer import ForkingMixIn
-try:    from http.server import HTTPServer
-except: from BaseHTTPServer import HTTPServer
+import signal, socket, optparse, time, os, sys, subprocess, logging, errno, ssl
+try:
+    from socketserver import ThreadingMixIn
+except ImportError:
+    from SocketServer import ThreadingMixIn
+
+try:
+    from http.server import HTTPServer
+except ImportError:
+    from BaseHTTPServer import HTTPServer
+
 import select
-from websockify import websocket
+from websockify import websockifyserver
+from websockify import auth_plugins as auth
 try:
     from urllib.parse import parse_qs, urlparse
-except:
+except ImportError:
     from cgi import parse_qs
     from urlparse import urlparse
 
-class ProxyRequestHandler(websocket.WebSocketRequestHandler):
+class ProxyRequestHandler(websockifyserver.WebSockifyRequestHandler):
+
+    buffer_size = 65536
 
     traffic_legend = """
 Traffic Legend:
@@ -37,20 +46,58 @@ Traffic Legend:
     <  - Client send
     <. - Client send partial
 """
+    
+    def send_auth_error(self, ex):
+        self.send_response(ex.code, ex.msg)
+        self.send_header('Content-Type', 'text/html')
+        for name, val in ex.headers.items():
+            self.send_header(name, val)
+        
+        self.end_headers()
+    
+    def validate_connection(self):
+        if not self.server.token_plugin:
+            return
+
+        host, port = self.get_target(self.server.token_plugin)
+        if host == 'unix_socket':
+            self.server.unix_target = port
+
+        else:
+            self.server.target_host = host
+            self.server.target_port = port
+
+    def auth_connection(self):
+        if not self.server.auth_plugin:
+            return
+
+        try:
+            # get client certificate data
+            client_cert_data = self.request.getpeercert()
+            # extract subject information
+            client_cert_subject = client_cert_data['subject']
+            # flatten data structure
+            client_cert_subject = dict([x[0] for x in client_cert_subject])
+            # add common name to headers (apache +StdEnvVars style)
+            self.headers['SSL_CLIENT_S_DN_CN'] = client_cert_subject['commonName']
+        except (TypeError, AttributeError, KeyError):
+            # not a SSL connection or client presented no certificate with valid data
+            pass
+            
+        try:
+            self.server.auth_plugin.authenticate(
+                headers=self.headers, target_host=self.server.target_host,
+                target_port=self.server.target_port)
+        except auth.AuthenticationError:
+            ex = sys.exc_info()[1]
+            self.send_auth_error(ex)
+            raise
 
     def new_websocket_client(self):
         """
         Called after a new WebSocket connection has been established.
         """
-        # Checks if we receive a token, and look
-        # for a valid target for it then
-        if self.server.token_plugin:
-            (self.server.target_host, self.server.target_port) = self.get_target(self.server.token_plugin, self.path)
-
-        if self.server.auth_plugin:
-            self.server.auth_plugin.authenticate(
-                headers=self.headers, target_host=self.server.target_host,
-                target_port=self.server.target_port)
+        # Checking for a token is done in validate_connection()
 
         # Connect to the target
         if self.server.wrap_cmd:
@@ -65,41 +112,66 @@ Traffic Legend:
             msg += " (using SSL)"
         self.log_message(msg)
 
-        tsock = websocket.WebSocketServer.socket(self.server.target_host,
-                                                 self.server.target_port,
-                connect=True, use_ssl=self.server.ssl_target, unix_socket=self.server.unix_target)
+        try:
+            tsock = websockifyserver.WebSockifyServer.socket(self.server.target_host,
+                                                           self.server.target_port,
+                                                           connect=True,
+                                                           use_ssl=self.server.ssl_target,
+                                                           unix_socket=self.server.unix_target)
+        except Exception:
+            self.log_message("Failed to connect to %s:%s",
+                             self.server.target_host, self.server.target_port)
+            raise self.CClose(1011, "Failed to connect to downstream server")
+
+        self.request.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
+        if not self.server.wrap_cmd and not self.server.unix_target:
+            tsock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
 
         self.print_traffic(self.traffic_legend)
 
         # Start proxying
         try:
             self.do_proxy(tsock)
-        except:
+        finally:
             if tsock:
                 tsock.shutdown(socket.SHUT_RDWR)
                 tsock.close()
                 if self.verbose:
                     self.log_message("%s:%s: Closed target",
                             self.server.target_host, self.server.target_port)
-            raise
 
-    def get_target(self, target_plugin, path):
+    def get_target(self, target_plugin):
         """
-        Parses the path, extracts a token, and looks up a target
-        for that token using the token plugin. Sets
-        target_host and target_port if successful
+        Gets a token from either the path or the host,
+        depending on --host-token, and looks up a target
+        for that token using the token plugin. Used by
+        validate_connection() to set target_host and target_port.
         """
         # The files in targets contain the lines
         # in the form of token: host:port
 
-        # Extract the token parameter from url
-        args = parse_qs(urlparse(path)[4]) # 4 is the query from url
+        if self.host_token:
+            # Use hostname as token
+            token = self.headers.get('Host')
 
-        if not 'token' in args or not len(args['token']):
+            # Remove port from hostname, as it'll always be the one where
+            # websockify listens (unless something between the client and
+            # websockify is redirecting traffic, but that's beside the point)
+            if token:
+                token = token.partition(':')[0]
+
+        else:
+            # Extract the token parameter from url
+            args = parse_qs(urlparse(self.path)[4]) # 4 is the query from url
+
+            if 'token' in args and len(args['token']):
+                token = args['token'][0].rstrip('\n')
+            else:
+                token = None
+
+        if token is None:
             raise self.server.EClose("Token not present")
 
-        token = args['token'][0].rstrip('\n')
-
         result_pair = target_plugin.lookup(token)
 
         if result_pair is not None:
@@ -192,12 +264,10 @@ Traffic Legend:
                 cqueue.append(buf)
                 self.print_traffic("{")
 
-class WebSocketProxy(websocket.WebSocketServer):
+class WebSocketProxy(websockifyserver.WebSockifyServer):
     """
     Proxy traffic to and from a WebSockets client to a normal TCP
-    socket server target. All traffic to/from the client is base64
-    encoded/decoded to allow binary data to be sent/received to/from
-    the target.
+    socket server target.
     """
 
     buffer_size = 65536
@@ -213,6 +283,7 @@ class WebSocketProxy(websocket.WebSocketServer):
         self.heartbeat      = kwargs.pop('heartbeat', None)
 
         self.token_plugin = kwargs.pop('token_plugin', None)
+        self.host_token = kwargs.pop('host_token', None)
         self.auth_plugin = kwargs.pop('auth_plugin', None)
 
         # Last 3 timestamps command was run
@@ -247,7 +318,7 @@ class WebSocketProxy(websocket.WebSocketServer):
                 "REBIND_OLD_PORT": str(kwargs['listen_port']),
                 "REBIND_NEW_PORT": str(self.target_port)})
 
-        websocket.WebSocketServer.__init__(self, RequestHandlerClass, *args, **kwargs)
+        websockifyserver.WebSockifyServer.__init__(self, RequestHandlerClass, *args, **kwargs)
 
     def run_wrap_cmd(self):
         self.msg("Starting '%s'", " ".join(self.wrap_cmd))
@@ -270,12 +341,17 @@ class WebSocketProxy(websocket.WebSocketServer):
         else:
             dst_string = "%s:%s" % (self.target_host, self.target_port)
 
-        if self.token_plugin:
-            msg = "  - proxying from %s:%s to targets generated by %s" % (
-                self.listen_host, self.listen_port, type(self.token_plugin).__name__)
+        if self.listen_fd != None:
+            src_string = "inetd"
         else:
-            msg = "  - proxying from %s:%s to %s" % (
-                self.listen_host, self.listen_port, dst_string)
+            src_string = "%s:%s" % (self.listen_host, self.listen_port)
+
+        if self.token_plugin:
+            msg = "  - proxying from %s to targets generated by %s" % (
+                src_string, type(self.token_plugin).__name__)
+        else:
+            msg = "  - proxying from %s to %s" % (
+                src_string, dst_string)
 
         if self.ssl_target:
             msg += " (using SSL)"
@@ -318,19 +394,65 @@ def _subprocess_setup():
     signal.signal(signal.SIGPIPE, signal.SIG_DFL)
 
 
-def logger_init():
+try :
+    # First try SSL options for Python 3.4 and above
+    SSL_OPTIONS = {
+        'default': ssl.OP_ALL,
+        'tlsv1_1': ssl.PROTOCOL_TLS | ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 |
+        ssl.OP_NO_TLSv1,
+        'tlsv1_2': ssl.PROTOCOL_TLS | ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 |
+        ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1,
+        'tlsv1_3': ssl.PROTOCOL_TLS | ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 |
+        ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_TLSv1_2,
+    }
+except AttributeError:
+    try:
+        # Python 3.3 uses a different scheme for SSL options
+        # tlsv1_3 is not supported on older Python versions
+        SSL_OPTIONS = {
+            'default': ssl.OP_ALL,
+            'tlsv1_1': ssl.PROTOCOL_TLSv1 | ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 |
+            ssl.OP_NO_TLSv1,
+            'tlsv1_2': ssl.PROTOCOL_TLSv1 | ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 |
+            ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1,
+        }
+    except AttributeError:
+        # Python 2.6 does not support TLS v1.2, and uses a different scheme
+        # for SSL options
+        SSL_OPTIONS = {
+            'default': ssl.PROTOCOL_SSLv23,
+            'tlsv1_1': ssl.PROTOCOL_TLSv1,
+        }
+
+def select_ssl_version(version):
+    """Returns SSL options for the most secure TSL version available on this
+    Python version"""
+    if version in SSL_OPTIONS:
+        return SSL_OPTIONS[version]
+    else:
+        # It so happens that version names sorted lexicographically form a list
+        # from the least to the most secure
+        keys = list(SSL_OPTIONS.keys())
+        keys.sort() 
+        fallback = keys[-1]
+        logger = logging.getLogger(WebSocketProxy.log_prefix)
+        logger.warn("TLS version %s unsupported. Falling back to %s",
+                    version, fallback)
+
+        return SSL_OPTIONS[fallback]
+
+def websockify_init():
+    # Setup basic logging to stderr.
     logger = logging.getLogger(WebSocketProxy.log_prefix)
     logger.propagate = False
     logger.setLevel(logging.INFO)
-    h = logging.StreamHandler()
-    h.setLevel(logging.DEBUG)
-    h.setFormatter(logging.Formatter("%(message)s"))
-    logger.addHandler(h)
-
-
-def websockify_init():
-    logger_init()
+    stderr_handler = logging.StreamHandler()
+    stderr_handler.setLevel(logging.DEBUG)
+    log_formatter = logging.Formatter("%(message)s")
+    stderr_handler.setFormatter(log_formatter)
+    logger.addHandler(stderr_handler)
 
+    # Setup optparse.
     usage = "\n    %prog [options]"
     usage += " [source_addr:]source_port [target_addr:target_port]"
     usage += "\n    %prog [options]"
@@ -356,14 +478,33 @@ def websockify_init():
             help="SSL certificate file")
     parser.add_option("--key", default=None,
             help="SSL key file (if separate from cert)")
+    parser.add_option("--key-password", default=None,
+            help="SSL key password")
     parser.add_option("--ssl-only", action="store_true",
             help="disallow non-encrypted client connections")
     parser.add_option("--ssl-target", action="store_true",
             help="connect to SSL target as SSL client")
+    parser.add_option("--verify-client", action="store_true",
+            help="require encrypted client to present a valid certificate "
+            "(needs Python 2.7.9 or newer or Python 3.4 or newer)")
+    parser.add_option("--cafile", metavar="FILE",
+            help="file of concatenated certificates of authorities trusted "
+            "for validating clients (only effective with --verify-client). "
+            "If omitted, system default list of CAs is used.")
+    parser.add_option("--ssl-version", type="choice", default="default",
+            choices=["default", "tlsv1_1", "tlsv1_2", "tlsv1_3"], action="store",
+            help="minimum TLS version to use (default, tlsv1_1, tlsv1_2, tlsv1_3)")
+    parser.add_option("--ssl-ciphers", action="store",
+            help="list of ciphers allowed for connection. For a list of "
+            "supported ciphers run `openssl ciphers`")
     parser.add_option("--unix-target",
             help="connect to unix socket target", metavar="FILE")
+    parser.add_option("--inetd",
+            help="inetd mode, receive listening socket from stdin", action="store_true")
     parser.add_option("--web", default=None, metavar="DIR",
             help="run webserver on same port. Serve files from DIR.")
+    parser.add_option("--web-auth", action="store_true",
+            help="require authentication to access webserver.")
     parser.add_option("--wrap-mode", default="exit", metavar="MODE",
             choices=["exit", "ignore", "respawn"],
             help="action to take when the wrapped program exits "
@@ -380,34 +521,108 @@ def websockify_init():
             "directory containing configuration files of this form "
             "(DEPRECATED: use `--token-plugin TokenFile --token-source "
             " path/to/token/file` instead)")
-    parser.add_option("--token-plugin", default=None, metavar="PLUGIN",
-                      help="use the given Python class to process tokens "
-                           "into host:port pairs")
+    parser.add_option("--token-plugin", default=None, metavar="CLASS",
+                      help="use a Python class, usually one from websockify.token_plugins, "
+                           "such as TokenFile, to process tokens into host:port pairs")
     parser.add_option("--token-source", default=None, metavar="ARG",
-                      help="an argument to be passed to the token plugin"
+                      help="an argument to be passed to the token plugin "
                            "on instantiation")
-    parser.add_option("--auth-plugin", default=None, metavar="PLUGIN",
-                      help="use the given Python class to determine if "
-                           "a connection is allowed")
+    parser.add_option("--host-token", action="store_true",
+                      help="use the host HTTP header as token instead of the "
+                           "token URL query parameter")
+    parser.add_option("--auth-plugin", default=None, metavar="CLASS",
+                      help="use a Python class, usually one from websockify.auth_plugins, "
+                           "such as BasicHTTPAuth, to determine if a connection is allowed")
     parser.add_option("--auth-source", default=None, metavar="ARG",
-                      help="an argument to be passed to the auth plugin"
+                      help="an argument to be passed to the auth plugin "
                            "on instantiation")
-    parser.add_option("--auto-pong", action="store_true",
-            help="Automatically respond to ping frames with a pong")
-    parser.add_option("--heartbeat", type=int, default=0,
-            help="send a ping to the client every HEARTBEAT seconds")
+    parser.add_option("--heartbeat", type=int, default=0, metavar="INTERVAL",
+            help="send a ping to the client every INTERVAL seconds")
+    parser.add_option("--log-file", metavar="FILE",
+            dest="log_file",
+            help="File where logs will be saved")
+    parser.add_option("--syslog", default=None, metavar="SERVER",
+            help="Log to syslog server. SERVER can be local socket, "
+                 "such as /dev/log, or a UDP host:port pair.")
+    parser.add_option("--legacy-syslog", action="store_true",
+                      help="Use the old syslog protocol instead of RFC 5424. "
+                           "Use this if the messages produced by websockify seem abnormal.")
 
     (opts, args) = parser.parse_args()
 
-    if opts.verbose:
-        logging.getLogger(WebSocketProxy.log_prefix).setLevel(logging.DEBUG)
+
+    # Validate options.
 
     if opts.token_source and not opts.token_plugin:
         parser.error("You must use --token-plugin to use --token-source")
 
+    if opts.host_token and not opts.token_plugin:
+        parser.error("You must use --token-plugin to use --host-token")
+
     if opts.auth_source and not opts.auth_plugin:
         parser.error("You must use --auth-plugin to use --auth-source")
 
+    if opts.web_auth and not opts.auth_plugin:
+        parser.error("You must use --auth-plugin to use --web-auth")
+
+    if opts.web_auth and not opts.web:
+        parser.error("You must use --web to use --web-auth")
+
+    if opts.legacy_syslog and not opts.syslog:
+        parser.error("You must use --syslog to use --legacy-syslog")
+
+
+    opts.ssl_options = select_ssl_version(opts.ssl_version)
+    del opts.ssl_version
+
+
+    if opts.log_file:
+        # Setup logging to user-specified file.
+        opts.log_file = os.path.abspath(opts.log_file)
+        log_file_handler = logging.FileHandler(opts.log_file)
+        log_file_handler.setLevel(logging.DEBUG)
+        log_file_handler.setFormatter(log_formatter)
+        logger.addHandler(log_file_handler)
+
+    del opts.log_file
+
+    if opts.syslog:
+        # Determine how to connect to syslog...
+        if opts.syslog.count(':'):
+            # User supplied a host:port pair.
+            syslog_host, syslog_port = opts.syslog.rsplit(':', 1)
+            try:
+                syslog_port = int(syslog_port)
+            except ValueError:
+                parser.error("Error parsing syslog port")
+            syslog_dest = (syslog_host, syslog_port)
+        else:
+            # User supplied a local socket file.
+            syslog_dest = os.path.abspath(opts.syslog)
+
+        from websockify.sysloghandler import WebsockifySysLogHandler
+
+        # Determine syslog facility.
+        if opts.daemon:
+            syslog_facility = WebsockifySysLogHandler.LOG_DAEMON
+        else:
+            syslog_facility = WebsockifySysLogHandler.LOG_USER
+
+        # Start logging to syslog.
+        syslog_handler = WebsockifySysLogHandler(address=syslog_dest,
+                                                 facility=syslog_facility,
+                                                 ident='websockify',
+                                                 legacy=opts.legacy_syslog)
+        syslog_handler.setLevel(logging.DEBUG)
+        syslog_handler.setFormatter(log_formatter)
+        logger.addHandler(syslog_handler)
+
+    del opts.syslog
+    del opts.legacy_syslog
+
+    if opts.verbose:
+        logger.setLevel(logging.DEBUG)
+
 
     # Transform to absolute path as daemon may chdir
     if opts.target_cfg:
@@ -419,43 +634,57 @@ def websockify_init():
 
     del opts.target_cfg
 
-    # Sanity checks
-    if len(args) < 2 and not (opts.token_plugin or opts.unix_target):
-        parser.error("Too few arguments")
     if sys.argv.count('--'):
         opts.wrap_cmd = args[1:]
     else:
         opts.wrap_cmd = None
-        if len(args) > 2:
-            parser.error("Too many arguments")
 
-    if not websocket.ssl and opts.ssl_target:
+    if not websockifyserver.ssl and opts.ssl_target:
         parser.error("SSL target requested and Python SSL module not loaded.");
 
     if opts.ssl_only and not os.path.exists(opts.cert):
         parser.error("SSL only and %s not found" % opts.cert)
 
-    # Parse host:port and convert ports to numbers
-    if args[0].count(':') > 0:
-        opts.listen_host, opts.listen_port = args[0].rsplit(':', 1)
-        opts.listen_host = opts.listen_host.strip('[]')
+    if opts.inetd:
+        opts.listen_fd = sys.stdin.fileno()
     else:
-        opts.listen_host, opts.listen_port = '', args[0]
+        if len(args) < 1:
+            parser.error("Too few arguments")
+        arg = args.pop(0)
+        # Parse host:port and convert ports to numbers
+        if arg.count(':') > 0:
+            opts.listen_host, opts.listen_port = arg.rsplit(':', 1)
+            opts.listen_host = opts.listen_host.strip('[]')
+        else:
+            opts.listen_host, opts.listen_port = '', arg
 
-    try:    opts.listen_port = int(opts.listen_port)
-    except: parser.error("Error parsing listen port")
+        try:
+            opts.listen_port = int(opts.listen_port)
+        except ValueError:
+            parser.error("Error parsing listen port")
+
+    del opts.inetd
 
     if opts.wrap_cmd or opts.unix_target or opts.token_plugin:
         opts.target_host = None
         opts.target_port = None
     else:
-        if args[1].count(':') > 0:
-            opts.target_host, opts.target_port = args[1].rsplit(':', 1)
+        if len(args) < 1:
+            parser.error("Too few arguments")
+        arg = args.pop(0)
+        if arg.count(':') > 0:
+            opts.target_host, opts.target_port = arg.rsplit(':', 1)
             opts.target_host = opts.target_host.strip('[]')
         else:
             parser.error("Error parsing target")
-        try:    opts.target_port = int(opts.target_port)
-        except: parser.error("Error parsing target port")
+
+        try:
+            opts.target_port = int(opts.target_port)
+        except ValueError:
+            parser.error("Error parsing target port")
+
+    if len(args) > 0 and opts.wrap_cmd == None:
+        parser.error("Too many arguments")
 
     if opts.token_plugin is not None:
         if '.' not in opts.token_plugin:
@@ -497,7 +726,7 @@ def websockify_init():
         server.start_server()
 
 
-class LibProxyServer(ForkingMixIn, HTTPServer):
+class LibProxyServer(ThreadingMixIn, HTTPServer):
     """
     Just like WebSocketProxy, but uses standard Python SocketServer
     framework.
@@ -546,7 +775,7 @@ class LibProxyServer(ForkingMixIn, HTTPServer):
     def process_request(self, request, client_address):
         """Override process_request to implement a counter"""
         self.handler_id += 1
-        ForkingMixIn.process_request(self, request, client_address)
+        ThreadingMixIn.process_request(self, request, client_address)
 
 
 if __name__ == '__main__':
